{------------------------------------------------------------------------------}
{   CRC32.pas                                                                  }
{                                                                              }
{   Copyright (c) 2001 by Pascal de Bruijn                                     }
{   All rights reserved worldwide                                              }
{                                                                              }
{   JSourcEd is distributed on a GPL license                                   }
{   (http://www.gnu.org/copyleft/gpl.html).                                    }
{                                                                              }
{------------------------------------------------------------------------------}

unit CRC32;

interface

  procedure InitCRC32(var CRC32: LongInt);
  procedure UpdateCRC32(var CRC32: LongInt; Data: Byte);
  procedure FinishCRC32(var CRC32: LongInt);

implementation

const
  Polynomial        = $EDB88320;

var
  CRC32Table        : array[0..255] of LongInt;

{------------------------------------------------------------------------------}

procedure BuildCRC32Table;
var
  I, J           : Byte;
begin
  for I := 0 to 255 do
  begin
    CRC32Table[I] := I;
    for J := 1 to 8 do
    begin
      if Odd(CRC32Table[I]) then CRC32Table[I] := (CRC32Table[I] shr 1) xor Polynomial
      else CRC32Table[I] := CRC32Table[I] shr 1;
    end;
  end;
end; { InitCRC32Table }

{------------------------------------------------------------------------------}

procedure InitCRC32(var CRC32: LongInt);
begin
  CRC32 := $FFFFFFFF;
end; { InitCRC32 }

{------------------------------------------------------------------------------}

procedure UpdateCRC32(var CRC32: LongInt; Data: Byte);
begin
  CRC32 := CRC32Table[Byte(CRC32 xor LongInt(Data))] xor ((CRC32 shr 8) and $00FFFFFF);
end; { UpdateCRC32 }

{------------------------------------------------------------------------------}

procedure FinishCRC32(var CRC32: LongInt);
begin
  CRC32 := CRC32 xor $FFFFFFFF;
end; { FinishCRC32 }

{------------------------------------------------------------------------------}

initialization
  BuildCRC32Table;
end. { CRC32 }