{------------------------------------------------------------------------------}
{   DosProcess.pas                                                             }
{                                                                              }
{   Copyright (c) 2001 by Pascal de Bruijn                                     }
{                                                                              }
{   JSourcEd is distributed on a GPL license                                   }
{   (http://www.gnu.org/copyleft/gpl.html).                                    }
{                                                                              }
{   All rights reserved worldwide                                              }
{------------------------------------------------------------------------------}

unit DosProcess;

{------------------------------------------------------------------------------}
interface
{------------------------------------------------------------------------------}

uses SysUtils, Windows;

type
  EDosProcessError = class(Exception);

function CaptureConsole(CommandLine, TmpFile: string; ShowWindow, ReturnStd: Boolean): String;

function CreateChildProcess(CommandLine: string; ShowWindow: Boolean; hStdIn, hStdOut, hStdErr: THandle): boolean;

const
  DUPLICATE_CLOSE_SOURCE = $00000001;
  DUPLICATE_SAME_ACCESS  = $00000002;
  BUFSIZE                = 4096;

{------------------------------------------------------------------------------}
implementation
{------------------------------------------------------------------------------}

procedure WriteToPipe(hStdInWr, hStdInRd: THandle);
var
  dwRead, dwWritten             : DWord;
  chBuf                         : Array [0..BUFSIZE] of CHAR;
begin
  while True do begin
    if (not ReadFile(hStdInRd, chBuf, BUFSIZE, dwRead, nil) or (dwRead = 0)) then Break;
    if (not WriteFile(hStdinWr, chBuf, dwRead, dwWritten, nil)) then Break;
  end;

  if (not CloseHandle(hStdinWr)) then raise EDosProcessError.Create('Close pipe failed');
end; { WriteToPipe }

{------------------------------------------------------------------------------}

function ReadFromPipe(hStdOutWr, hStdOutRd, hStdOutSave, hDataFile: THandle): string;
var
  dwRead, dwWritten              : DWord;
  chBuf                          : Array [0..BUFSIZE] of CHAR;
  ReadOK                         : Boolean;
  ReadMore                       : Boolean;
  NewStr                         : String;
begin
  if (not CloseHandle(hStdoutWr)) then raise EDosProcessError.Create('Closing handle failed');

  repeat
    ReadOK := ReadFile(hStdOutRd, chBuf, BUFSIZE, dwRead, nil);
    NewStr := NewStr + string(chBuf);
    ReadMore := ReadOK and (dwRead <> 0);
    if ReadMore then ReadMore := WriteFile(hDataFile, chBuf, dwRead, dwWritten, nil);
  until (not ReadMore);
  Result := NewStr;
end; { ReadFromPipe }

{------------------------------------------------------------------------------}

function CaptureConsole(CommandLine, TmpFile: string; ShowWindow, ReturnStd: Boolean): String;
var
  saAttr                               : TSECURITYATTRIBUTES;
  fSuccess                             : Boolean;
  hChildStdinRd, hChildStdinWr         : THandle;
  hChildStdinWrDup, hChildStdoutRd     : THandle;
  hChildStdoutWr                       : THandle;
  hSaveStdin, hSaveStdout,hSaveStdErr  : THandle;
  hInputFile                           : THandle;
begin
  saAttr.nLength := sizeof(TSecurityAttributes);
  saAttr.bInheritHandle := TRUE;
  saAttr.lpSecurityDescriptor := nil;

  hSaveStdin := GetStdHandle(STD_INPUT_HANDLE);
  hSaveStdout := GetStdHandle(STD_OUTPUT_HANDLE);
  hSaveStdErr := GetStdHandle(STD_ERROR_HANDLE);

  if (not CreatePipe(hChildStdoutRd, hChildStdoutWr, @saAttr, 0)) then raise EDosProcessError.Create('STD_OUT pipe creation failed');

  if ReturnStd then
  begin
    if (not SetStdHandle(STD_OUTPUT_HANDLE, hChildStdoutWr)) then raise EDosProcessError.Create('Redirecting STDOUT failed');
  end
  else
  begin
    if (not SetStdHandle(STD_ERROR_HANDLE, hChildStdoutWr)) then raise EDosProcessError.Create('Redirecting STD_ERR failed');
  end;

  if (not CreatePipe(hChildStdinRd, hChildStdinWr, @saAttr, 0)) then raise EDosProcessError.Create('STD_IN pipe creation failed');

  if (not SetStdHandle(STD_INPUT_HANDLE, hChildStdinRd)) then raise EDosProcessError.Create('Redirecting STD_IN failed');

  fSuccess := DuplicateHandle(GetCurrentProcess(), hChildStdinWr, GetCurrentProcess(), @hChildStdinWrDup, 0, FALSE, DUPLICATE_SAME_ACCESS);

  if (not fSuccess) then raise EDosProcessError.Create('DuplicateHandle failed');

  CloseHandle(hChildStdinWr);

  try
    if (not CreateChildProcess(CommandLine, ShowWindow,hChildStdinRd,hChildStdoutWr,hChildStdoutWr)) then raise EDosProcessError.Create('Create process failed');
  finally
    if (not SetStdHandle(STD_INPUT_HANDLE, hSaveStdin)) then raise EDosProcessError.Create('Re-redirecting STD_IN failed');
    if (not SetStdHandle(STD_OUTPUT_HANDLE, hSaveStdout)) then raise EDosProcessError.Create('Re-redirecting STD_OUT failed');
    if (not SetStdHandle(STD_ERROR_HANDLE, hSaveStdErr)) then raise EDosProcessError.Create('Re-redirecting STD_ERR failed');
  end;

  hInputFile := FileCreate(PChar(TmpFile));

  if (hInputFile = -1) then raise EDosProcessError.Create('No input file');
  try
    ReadFromPipe(hChildStdoutWr,hChildStdoutRd, hSaveStdout, hInputFile);
  finally
    FileClose(hInputFile);
  end;
  Result := TmpFile;
end; { CaptureConsole }

{------------------------------------------------------------------------------}

function CreateChildProcess(CommandLine: string; ShowWindow: Boolean; hStdIn, hStdOut, hStdErr: THandle): boolean;
var
  StartInfo: TSTARTUPINFO;
  ProcInfo: TProcessInformation;
begin
  GetStartupInfo(StartInfo);
  with StartInfo do
  begin
    cb := SizeOf(TStartupInfo);
    dwFlags := STARTF_USESHOWWINDOW or STARTF_FORCEONFEEDBACK or STARTF_USESTDHANDLES;
    hStdOutput := hStdout;
    hStdInput := hStdin;
    hStdError := hStdErr;
    if ShowWindow then wShowWindow := SW_SHOWNORMAL
    else wShowWindow := SW_SHOWMINNOACTIVE;
  end;

  Result :=  CreateProcess(nil, PChar(CommandLine), nil, nil, True, 0, nil, nil, StartInfo, ProcInfo);
end; { CreateChildProcess }

{------------------------------------------------------------------------------}

end. { DosProcess }
