
unit SnapToEdge;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs;

type
  TDirection = (dUp, dDown, dLeft, dRight, dNone);

  TSnapProc = procedure(Sender: TComponent; SnapPoint: TDirection) of object;

  TSnapToEdge = class(TComponent)
  private
    FOldWindowProc: TWndMethod;
    FOldRect: TRect;
    FHwnd: THandle;
    FWin: TCustomForm;
    HDir: TDirection;
    VDir: TDirection;
    FInnerBorderSize: Word;
    FOuterBorderSize: Word;
    FEnabled: Boolean;
    FonSnap: TSnapProc;
    procedure SetEnabled(const Value: Boolean);
    procedure SetInnerBorderSize(const Value: Word);
    procedure SetOuterBorderSize(const Value: Word);
    procedure SetOnSnap(const Value: TSnapProc);
    procedure DoSnap(Direction: TDirection);
  protected
    function Hook: Boolean;
    procedure Handle_Message(var message: TMessage);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    property Enabled: Boolean read FEnabled write SetEnabled;
    property InnerBorderSize: Word read FInnerBorderSize write SetInnerBorderSize;
    property OuterBorderSize: Word read FOuterBorderSize write SetOuterBorderSize;
    property OnSnap: TSnapProc read FOnSnap write SetOnSnap;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Misc', [TSnapToEdge]);
end; { Register }

constructor TSnapToEdge.Create(AOwner: TComponent);
var
  F: TCustomForm absolute AOwner;
begin
  inherited;
  FInnerBorderSize := 16;
  FOuterBorderSize := 4;
  Enabled := True;
  if Assigned(AOwner) then
  begin
    if AOwner is TCustomForm then
    begin
      FWin := F;
      FHwnd := F.Handle;
      Hook;
    end
    else
    begin
      // Cannot hook other components
    end;
  end;
end; { Create }

function TSnapToEdge.Hook: Boolean;
begin
  FOldWindowProc := FWin.WindowProc;
  FWin.WindowProc := Handle_Message;
  Result := True;
end; { Hook }

procedure TSnapToEdge.DoSnap(Direction: TDirection);
begin
  if Assigned(FOnSnap) then FOnSnap(Self, Direction);
end; { DoSnap }

procedure TSnapToEdge.Handle_Message(var message: TMessage);
var
  MsgRect                                : PRect;
  NewRect                                : TRect;
  OldRect                                : TRect;
  WindowHeight                           : Integer;
  WindowWidth                            : Integer;
  DesktopRect                            : TRect;
begin
  if (Enabled) and (message.Msg = WM_MOVING) then
  begin
    MsgRect := Pointer(message.lParam);
    if Assigned(MsgRect) then
    begin
      NewRect := MsgRect^;
      OldRect := FOldRect;

      if NewRect.Top < OldRect.Top then VDir := dUp
      else
      if NewRect.Top > OldRect.Top then VDir := dDown
      else VDir := dNone;

      if NewRect.Left < OldRect.Left then HDir := dLeft
      else
      if NewRect.Left > OldRect.Left then HDir := dRight
      else HDir := dNone;

      case VDir of
        dUp:
        begin
          if (NewRect.Top <= FInnerBorderSize) and (NewRect.Top + FOuterBorderSize > 0) then
          begin
            NewRect.Bottom := NewRect.Bottom - NewRect.Top;
            NewRect.Top := 0;
            MsgRect^.Top := NewRect.Top;
            MsgRect^.Bottom := NewRect.Bottom;
            DoSnap(dUp);
          end;
        end;
        dDown:
        begin
          if not GetClientRect(GetDeskTopWindow, DesktopRect) then
          begin
            DesktopRect := Rect(0, 0, Screen.Height, Screen.Width);
          end;
          if (NewRect.Bottom + FInnerBorderSize >= Screen.Height) and (NewRect.Bottom < Screen.Height + FOuterBorderSize) then
          begin
            WindowHeight := NewRect.Bottom - NewRect.Top;
            NewRect.Bottom := DesktopRect.Bottom;
            NewRect.Top := NewRect.Bottom - WindowHeight;
            MsgRect^.Top := NewRect.Top;
            MsgRect^.Bottom := NewRect.Bottom;
            DoSnap(dDown);
          end;
        end;
      end;

      case HDir of
        dLeft:
        begin
          if (NewRect.Left < FInnerBorderSize) and (NewRect.Left + OuterBorderSize > 0) then
          begin
            NewRect.Right := NewRect.Right - NewRect.Left;
            NewRect.Left := 0;
            MsgRect^.Left := NewRect.Left;
            MsgRect^.Right := NewRect.Right;
            DoSnap(dLeft);
          end;
        end;
        dRight:
        begin
          if ((NewRect.Right + FInnerBorderSize) > Screen.Width) and (NewRect.Right < Screen.Width + FOuterBorderSize) then
          begin
            WindowWidth := NewRect.Right - NewRect.Left;
            NewRect.Right := Screen.Width;
            NewRect.Left := Screen.Width - WindowWidth;
            MsgRect^.Left := NewRect.Left;
            MsgRect^.Right := NewRect.Right;
            DoSnap(dRight);
          end;
        end;
      end;

      FOldRect := MsgRect^;
    end;
  end
  else
  begin
    if Assigned(FOldWindowProc) then FOldWindowProc(message);
  end;
end; { Handle_Message }

procedure TSnapToEdge.SetEnabled(const Value: Boolean);
begin
  FEnabled := Value;
end; { SetEnabled }

procedure TSnapToEdge.SetInnerBorderSize(const Value: Word);
begin
  FInnerBorderSize := Value;
end; { SetInnerBorderSize }

procedure TSnapToEdge.SetOuterBorderSize(const Value: Word);
begin
  FOuterBorderSize := Value;
end; { SetOuterBorderSize }

procedure TSnapToEdge.SetOnSnap(const Value: TSnapProc);
begin
  FOnSnap := Value;
end; { SetOnSnap }

destructor TSnapToEdge.Destroy;
begin
  FWin.WindowProc := FOldWindowProc;
  inherited;
end; { Destroy }

end.
