{*************************************************************}
{            Shell Extension Component for Delphi 32          }
{ Version:   1.36                                             }
{ E-Mail:    info@utilmind.com                                }
{ Home Page: www.utilmind.com                                 }
{ Created:   April 15, 1999                                   }
{ Modified:  June 10, 2000                                    }
{ Legal:     Copyright (c) 1999-2000, UtilMind Solutions      }
{*************************************************************}
{   TShellExt                                                 }
{ Component for determination of the information associated   }
{ with file extensions (its executable application, icons     }
{ (32x32 and 16x16), description) and/or for installing/      }
{ uninstalling of the file extensions to the OS.              }
{ This component works with extensions through the System     }
{ Registry.                                                   }
{*************************************************************}
{ PROPERTIES:                                                 }
{   LargeIcon: TIcon - 32x32 icon of Shell Extension (Result  }
{                 of GetShellExtension routine, EXTRACT ONLY) }
{   SmallIcon: TIcon - 16x16, Small Icon of ShellExtension    }
{                     (result of GetShellExtension routine)   }
{                                                             }
{   IconIndex: Integer -  Used for INSTALL ONLY               }
{                         (icon index in .exe or .dll)        }
{   ExtDescription: String - Description of Extension         }
{                            uses in System Registry          }
{   FileDescription: String - Description of File Type        }
{   Extension: String - File extension (.txt or .pas erc)     }
{   OpenWith: String - Extension opens with this [filename]   }
{   OpenWithMe: Boolean - Open with current .exe project      }
{   ParamString: String - Run parameters. Usualy '%1'.        }
{ METHODS:                                                    }
{   Install - installs custom Shell-extension (icon must be   }
{             specified by IconIndex property)                }
{   UnInstall - unistalls Shell-extension                     }
{   GetShellExtension - extract ShellExtension into component }
{                       Icon property will be extension icon. }
{*************************************************************}
{ Please see demo program for more information.               }
{*************************************************************}
{                     IMPORTANT NOTE:                         }
{ This software is provided 'as-is', without any express or   }
{ implied warranty. In no event will the author be held       }
{ liable for any damages arising from the use of this         }
{ software.                                                   }
{ Permission is granted to anyone to use this software for    }
{ any purpose, including commercial applications, and to      }
{ alter it and redistribute it freely, subject to the         }
{ following restrictions:                                     }
{ 1. The origin of this software must not be misrepresented,  }
{    you must not claim that you wrote the original software. }
{    If you use this software in a product, an acknowledgment }
{    in the product documentation would be appreciated but is }
{    not required.                                            }
{ 2. Altered source versions must be plainly marked as such,  }
{    and must not be misrepresented as being the original     }
{    software.                                                }
{ 3. This notice may not be removed or altered from any       }
{    source distribution.                                     }
{*************************************************************}

unit ShellExt;

interface

uses
  Windows, Classes, Graphics, Controls, Forms, SysUtils, Registry,
  ShellAPI;

type
  TShellExt = class(TComponent)
  private
    FLargeIcon: TIcon;   { Used for GetShellExtension only }
    FSmallIcon: TIcon;   { Used for GetShellExtension only }
    FExtension: String;
    FExtDescription: String; { For example "UtilMind scheme file" }
    FFileDescription: String; { Scheme file }
    FOpenWith: String; { For example 'D:\PROGRAM\PROGRAM.EXE' (file will be opens as [program.exe file.sh]) }
    FOpenWithMe: Boolean; { Open with current .exe file }
    FParamString: String; { For example %1 }
    FIconIndex: Integer; { Used for Install only (icon index in .exe or .dll) }
    procedure SetLargeIcon(Value: TIcon);
    procedure SetSmallIcon(Value: TIcon);
    procedure SetExtension(Value: String);
    procedure SetOpenWith(Value: String);
    procedure SetOpenWithMe(Value: Boolean);
  public
    constructor Create(aOwner: TComponent); override;
    destructor Destroy; override;

    function Install: Boolean; { True if successful }
    function UnInstall: Boolean;
    function GetShellExtension: Boolean;
  published
    { See description above }
    property LargeIcon: TIcon read FLargeIcon write SetLargeIcon;
    property SmallIcon: TIcon read FSmallIcon write SetSmallIcon;
    property IconIndex: Integer read FIconIndex write FIconIndex;
    property ExtDescription: String read FExtDescription write FExtDescription;
    property FileDescription: String read FFileDescription write FFileDescription;
    property Extension: String read FExtension write SetExtension;
    property OpenWith: String read FOpenWith write SetOpenWith;
    property OpenWithMe: Boolean read FOpenWithMe write SetOpenWithMe;
    property ParamString: String read FParamString write FParamString;
  end;

procedure Register;

implementation

constructor TShellExt.Create(aOwner: TComponent);
begin
  inherited Create(aOwner);
  FOpenWith := '[Self]';
  FOpenWithMe := True;
  FLargeIcon := TIcon.Create;
  FSmallIcon := TIcon.Create;
end;

destructor TShellExt.Destroy;
begin
  FSmallIcon.Destroy;
  FLargeIcon.Destroy;
  inherited Destroy;
end;

procedure TShellExt.SetLargeIcon(Value: TIcon);
begin
  if FLargeIcon <> Value then FLargeIcon.Assign(Value);
end;

procedure TShellExt.SetSmallIcon(Value: TIcon);
begin
  if FSmallIcon <> Value then FSmallIcon.Assign(Value);
end;

procedure TShellExt.SetExtension(Value: String);
begin
  if FExtension <> Value then
   FExtension := Value;
end;

procedure TShellExt.SetOpenWith(Value: String);
begin
  if not FOpenWithMe and (FOpenWith <> Value) then
    FOpenWith := Value;
end;

procedure TShellExt.SetOpenWithMe(Value: Boolean);
begin
  FOpenWithMe := Value;
  if Value then
   if not (csDesigning in ComponentState) then
    FOpenWith := Application.ExeName
   else FOpenWith := '[Self]';
end;

function TShellExt.Install: Boolean;
var
  Reg: TRegistry;
begin
  Reg := TRegistry.Create;
  try
    with Reg do
     begin
      RootKey := HKEY_CLASSES_ROOT;
      OpenKey(FExtension, True);
      WriteString('', FExtDescription);
      OpenKey('\' + FExtDescription, True);
      WriteString('', FFileDescription);
      OpenKey('DefaultIcon', True);
      WriteString('', OpenWith + ','+IntToStr(FIconIndex));
      OpenKey('\' + FExtDescription + '\Shell\Open\Command', True);
      WriteString('', '"' + OpenWith + '" "' + FParamString + '"');
     end;
   Result := True;  
  except
   Result := False;
  end;
  Reg.Free;
end;

function TShellExt.UnInstall: Boolean;
var
  Reg: TRegistry;
begin
  Reg := TRegistry.Create;
  try
    with Reg do
     begin
      RootKey := HKEY_CLASSES_ROOT;
      DeleteKey(FExtension);
      DeleteKey('\' + FExtDescription);
     end;
   Result := True;
  except
   Result := False;
  end;
  Reg.Free;
end;

function TShellExt.GetShellExtension: Boolean;
var
  LIcon, SIcon: hIcon;
  Reg: TRegistry;
  IconFileName, IconIndex: String;
  PC: Array[0..$FF] of Char;
  i, j: Integer;

  procedure SplitStr(SplitChar: String; var Str, Str1: String);
  var
    i: Integer;
  begin
    i := Pos(SplitChar, Str);
    if i <> 0 then
     begin
      Str1 := Copy(Str, i + 1, Length(Str) - i + 1);
      SetLength(Str, i - 1);
     end
    else FParamString := '';
  end;

begin
  Reg := TRegistry.Create;
  try
    with Reg do
     begin
      RootKey := HKEY_CLASSES_ROOT;
      OpenKey(FExtension, False);
      FExtDescription := ReadString('');
      OpenKey('\' + FExtDescription, False);
      FFileDescription := ReadString('');
      OpenKey('DefaultIcon', False);

      if (FExtDescription = '') and (FExtDescription = '') then
       begin
        GetSystemDirectory(PC, $FF);
        StrCat(PC, '\SHELL32.DLL');

        ExtractIconEx(PC, 0, LIcon, SIcon, 1);
        FLargeIcon.Handle := LIcon;
        FSmallIcon.Handle := SIcon;

        Result := False;
        Exit;
       end;

      IconFileName := ReadString('');
      if IconFileName <> '' then
       begin
        SplitStr(',', IconFileName, IconIndex);
        StrPCopy(PC, IconFileName);

        try
          ExtractIconEx(PC, StrToInt(IconIndex), LIcon, SIcon, 1);
          FLargeIcon.Handle := LIcon;
          FSmallIcon.Handle := SIcon;
        except
          FLargeIcon.Handle := 0;
          FSmallIcon.Handle := 0;

          try
            GetSystemDirectory(PC, $FF);
            StrCat(PC, '\SHELL32.DLL');

            if LowerCase(FExtension) = '.exe' then i := 2
            else i := 0;
            ExtractIconEx(PC, i, LIcon, SIcon, 1);
            FLargeIcon.Handle := LIcon;
            FSmallIcon.Handle := SIcon;
          except
          end;
        end;
       end;

      OpenKey('\' + FExtDescription + '\Shell\Open\Command', False);
      FOpenWithMe := False;
      FOpenWith := ReadString('');

      if IconFileName = '' then
       begin
        IconFileName := FOpenWith;
        SplitStr(',', IconFileName, IconIndex);
        SplitStr(' -', IconFileName, IconIndex);        
        StrPCopy(PC, IconFileName);
        ExtractIconEx(PC, 0, LIcon, SIcon, 1);
        FLargeIcon.Handle := LIcon;
        FSmallIcon.Handle := SIcon;
       end;

      i := Pos('"', FOpenWith);
      if i = 1 then
       begin
        FOpenWith := Copy(FOpenWith, 2, Length(FOpenWith) - 1);
        i := Pos('"', FOpenWith);
        FParamString := Copy(FOpenWith, i + 2, Length(FOpenWith) - i - 1);
        j := Pos('"', FParamString);
        while j <> 0 do
         begin
          Delete(FParamString, j, 1);
          j := Pos('"', FParamString);
         end;
        FOpenWith := Copy(FOpenWith, 0, i - 1);
       end
      else SplitStr(' ', FOpenWith, FParamString)
     end;
    Result := True;
  except
   Result := False;
  end;
  Reg.Free;
end;

procedure Register;
begin
  RegisterComponents('UtilMind', [TShellExt]);
end;

end.
